
/* changes
 - 151123 (v1221): added trim() to lines reading filenames in readBatchFile(), and FinishPlugin() at the end of onStart()
 - 140623 (v122): experiment.applyFiles converted from object to string, updating splitting functions
 - 120623 (v122): readBatchFile: due to QDir and QFile cannot be used, readBatchFile() now activates the plugin to read the contents,
and processes the contents in the function
 - 060623 (v122): still converting, classes Experiment() line 587 and ProjectParameters line 602 and 957
 - 180423 (v122): convert *.js and *.ui to *.qml
 - 091222 (v121): adapt FMR and DMR format due to an update of FMR file version 7 (from 'ResolutionX' and 'ResolutionY' to 'NrOfColumns' and 'NrOfRows')
 - 090522 (v120): updated link for internal help (line 74)
 - 291021 (v120): 'apply to' files are printed to BV Log tab with empty lines, but this is required for plugin.
 - 121120 (v111): reset size of font in help fields to 8pt
 - 191219 (v111): added printing of parameters to BV Log tab
 - 141119 (v111): more restrictions to continue with applying VDMs in case of NaNs
 - 311019 (v111): fixed small bug where undistortion would be applied for option 0 (estimate only)
 - 181019 (v111): fixed small bug when using several apply files
 - 031019 (v111): removed 'the order does not matter'
 - 120718:
 - June 2018: added 'order' for Gaussian spinbox and help; added one tab in *.ui
 - June-July 2017: div.
 - 230517: added buttons for batch version
 - 230517: changed 'bvqx' to 'bv'
 - 230517: vdmBrowse(), and add() do not invoke the plugin anymore, but BrowseFile()
 - 230517: added batchFileBrowse(), startProcessing(), readBatchFile()
 - 290517: added ProjectParameters(), selectOption(), continued readBatchFile()
 - 300517: added Experiment(), closeAllProjects() from anatabacus
 */

import QtQml
import QtQuick
import QtQuick.Layouts
import QtQuick.Controls


Pane {
    id: root
    width: 1105
    height: 625

    TabBar {
        id: bar
        width: parent.width
        height: parent.height

        TabButton {
            text: qsTr("General options")
            //     height: implicitHeight
        }
        TabButton {
            text: qsTr("Parameters")
        }
        TabButton {
            text: qsTr("Single projects: estimate VDM")
        }
        TabButton {
            text: qsTr("Single projects: apply VDM")
        }
        //    TabButton {
        //        text: qsTr("Help")
        //    }
    }

    StackLayout {
        width: parent.width
        height: parent.height
        currentIndex: bar.currentIndex

        Item {
            id: generalOptsTab

            //ColumnLayout {
            GroupBox {
                id: grbxGeneralOptions
                title: "General options for image registration-based distortion Correction based on Opposite Phase Encoding (COPE)"
                Layout.fillWidth: true
                x: 20
                y: 40
                width: 931
                height: 551

                GroupBox {
                    id: grbx5
                    title: "File selection"
                    x: 100
                    y: 70
                    width: 791
                    height: 151

                    RadioButton {
                        id: rbtnSingleProject
                        text: qsTr("Single-project version: select files manually")
                        x: 8
                        y: 0
                        checked: true
                    }

                    RadioButton {
                        id: rbtnBatch
                        text: qsTr("Multiple-project version: load text file with filenames")
                        x: 8
                        y: 19
                        enabled: true
                    }

                    Row {
                        x: 30
                        spacing: 6
                    TextField {
                        id: batchFileLabel
                        font.pixelSize: 10
                        enabled: rbtnBatch.checked
                        placeholderText: "batch processing .txt file"
                        //x: 30
                        y: 48
                        width: 450
                    }
                    Button {
                        id: loadFileNamesButton
                        //x: 605
                        y: 47 // 60
                        text: "Load filenames..."
                        enabled: rbtnBatch.checked
                        onClicked: batchFileLabel.text = root.onGetFileName("*.txt")
                    }
                    }
                }

                GroupBox {
                    id: grbx2
                    title: "Please select a processing option (single projects)"
                    x: 290
                    y: 250
                    width: 431
                    height: 131
                    enabled: rbtnSingleProject.checked

                    RadioButton {
                        id: rbtnEstimateOnly
                        text: qsTr("Estimate voxel displacement map (VDM)")
                        x: 8
                        y: 4
                    }

                    RadioButton {
                        id: rbtnApplyOnly
                        text: qsTr("Apply voxel displacement map (VDM) to several runs")
                        x: 8
                        y: 27
                    }

                    RadioButton {
                        id: rbtnEstimateAndApply
                        checked: true
                        text: qsTr("Estimate VDM and apply VDM to same data")
                        x: 8
                        y: 50
                    }
                }
                //}
            }

            Row {
                x: 620
                y: grbxGeneralOptions.height - 20
                spacing: 5

                Item { Layout.fillWidth: true }
                Button {
                    id: cancelButton
                    text: "Cancel"
                    onClicked: BrainVoyager.FinishPlugin() // 'ClosePlugin' for pure qml scripts, but if qml dialog serves as frontend for plugin, 'FinishPlugin()' needs to be called!
                }
                Button {
                    id: startButton
                    text: "Start processing multiple projects"
                    onClicked: root.onStart()
                    enabled: true
                }
            }
        }
        Item {
            id: paramsTab

            GroupBox {
                id: grbxParams
                title: "Parameters for voxel displacement map (VDM) - all projects"
                x: 40
                y: 60
                width: 450
                height: 410

                GroupBox {
                    id: grbxParamsDistMeas
                    title: "Distance measure for local optimization"
                    Layout.fillWidth: true
                    x: 20
                    y: 30
                    width: 400
                    height: 80

                    RadioButton {
                        id: nccRadioButton
                        text: qsTr("Normalized Cross-Correlation (NCC)")
                        checked: true
                        x: 10
                        y: 0
                        width: 311
                        height: 20
                    }

                    RadioButton {
                        id: ssdRadioButton
                        text: qsTr("Sum of Squared Differences (SSD)")
                        x: 10
                        y: 20
                        width: 311
                        height: 20
                    }
                }
                GroupBox {
                    id: grbxParamsDataType
                    title: "Type of data"
                    x: 20
                    y: 110
                    width: 400
                    height: 80

                    RadioButton {
                        id: geRadioButton
                        text: qsTr("Gradient Echo (GE-EPI, commonly used for fMRI)")
                        checked: true
                        x: 10
                        y: 0
                        width: 300
                        height: 20
                    }

                    RadioButton {
                        id: seRadioButton
                        text: qsTr("Spin Echo (SE-EPI, commonly used for DTI)")
                        x: 10
                        y: 20
                        width: 300
                        height: 20
                    }
                }

                GroupBox {
                    id: grbxParamsPEDir
                    title: "Phase encoding direction"
                    width: 251
                    height: 80
                    x: 20
                    y: 190

                    RadioButton {
                        id: estimateAppaRadioButton
                        text: qsTr("Anterior (A) <--> Posterior (P)")
                        checked: true
                        x: 10
                        y: 0
                        width: 230
                        height: 20
                    }

                    RadioButton {
                        id: estimateRllrRadioButton
                        text: qsTr("Left (L) <--> Right (R)")
                        x: 10
                        y: 20
                        width: 311
                        height: 20
                    }

                }
                GroupBox {
                    id: grbxParamsGaussOrder
                    title: "Order of Gaussian"
                    x: 280
                    y: 190
                    width: 140
                    height: 80

                    SpinBox {
                        id: spbOrderGauss
                        Layout.maximumWidth: 60
                        from: 0
                        to: 2
                        value: 0
                        editable: true
                        x: 0
                        y: 10
                        width: 61
                        height: 24
                    }
                }
                GroupBox {
                    id: grbxParamsOtherOpts
                    title: "Other options"
                    Layout.fillWidth: true
                    x: 20
                    y: 270
                    width: 400
                    height: 80

                    CheckBox {
                        id: saveDerivative
                        text: qsTr("Save derivative of VDM")
                        x: 13 // position w.r.t. groupbox
                        y: 0
                        width: 191
                        height: 20
                        checked: false
                    }
                }

            } // end grbxParams
            TextArea {
                id: txtBrowser
                // text: placeholderText: "
                text: 
"**Type of data:** Select whether your data used for estimation is gradient-echo EPI (GE-EPI) or spin-echo EPI (SE-EPI)." +
"For spin-echo data the image intensities will be adjusted during optimization depending on the derivative of the " +
"Voxel Displacement Map (VDM), while for gradient-echo data it proved to be better to skip this adjustment. " +
"Gradient-echo is typically used for functional scans and for spin-echo for diffusion-weighted imaging (DWI). " +
"When it is unknown whether the data are spin-echo or gradient-echo, it is adviced to run the estimation once " +
"with each setting. The setting that provides the highest final NCC value (printed to the BrainVoyager Log tab) " +
"should be the setting that works best for the distortion correction." +
"\n\n" +
"**Distance measure for optimization:** This option offers the possibility to choose between two different distance measures" +
"used to estimate the similarity between the images:\n\n" +
"*normalised cross-correlation (NCC)* and *sum of squared differences (SSD)*." + // For more information, please see background." +
"\n\n" +
"**Apply VDM to input:** This option applies the voxel displacement map (VDM) to the data that have been used to estimate " +
"the distortion. In case there are no further runs, this eliminates the need to run the second step via the 'Apply' tab." +
"\n\n" +
"**Order of Gaussian**: This option sets the order of the Gaussian derivative function used to smooth during column-wise " +
"estimation of distortion. Default is 0, which is fine, except if the plugin reports not-a-number (NaN) values in the " +
"BrainVoyager Log tab, in this case one might want to change the order from 0 to for example 1. " +
"\n\n" +
"**Save derivative of VDM:** The second option is to save the derivative of the VDM. To view a map, open a corrected file " +
"(e.g., \"example_undist.fmr\"), go to Analysis > Overlay Volume Maps and load the corresponding map file " +
"(e.g. \"example_undist_vdmderiv.map\").\n" +
"The *_vdm map shows the voxel displacement map (see middle column in figure below); in areas marked in yellow/red the " +
"voxels have been moved up and in areas marked in green/blue the voxels have been moved down. " +
"The *_vdmderiv map contains the first derivative of the voxel displacements in the Y direction. " +
"This shows which areas have been compressed (yellow/red) or stretched (green/blue) during distortion correction."
                Layout.fillWidth: true
                x: 510
                y: 65
                width: 480
                height: 491
                font.pixelSize: Qt.platform.os === "osx" ? 12 : 10
                wrapMode: TextEdit.WordWrap
                textFormat: TextEdit.MarkdownText
                color: Qt.rgba(0.8, 0.8, 0.8, 1.0)
            }
        }
        Item {
            id: estimateVDMTab

            GroupBox {
                id: grbxEstimateVDM
                title: "Estimation of voxel displacement map (VDM): select files (for single project)"
                x: 10
                y: 50
                width: 1071
                height: 180

                Label {
                    text: "Select 2 FMRs/DMRs with opposite phase encoding directions and the specific volumes to use for the estimation:"
                    x: 1
                    y: 0
                    width: 813
                    height: 10
                }

                TextField {
                    id: filenameEdit0
                    x: 1
                    y: 30
                    width: 813
                    height: 21
                }
                Button {
                    id: browseButton0
                    text: "Browse..."
                    onClicked: filenameEdit0.text = root.onSelectInput(0)
                    x: 816
                    y: 30
                }
                Label {
                    text: "Volume:"
                    x: 910
                    y: 30
                }
                SpinBox {
                    id: volumeSpinBox0
                    Layout.maximumWidth: 60
                    from: 1
                    to: 1999
                    value: 1
                    editable: true
                    x: 977
                    y: 30
                    width: 64
                    height: 24
                }

                TextField {
                    id: filenameEdit1
                    x: 1
                    y: 55
                    width: 813
                    height: 21
                }
                Button {
                    id: browseButton1
                    text: "Browse..."
                    onClicked: filenameEdit1.text = root.onSelectInput(1)
                    x: 816
                    y: 55
                }
                Label {
                    text: "Volume:"
                    x: 910
                    y: 55
                }
                SpinBox {
                    id: volumeSpinBox1
                    Layout.maximumWidth: 60
                    from: 1
                    to: 1999
                    value: 1
                    editable: true
                    x: 977
                    y: 55
                    width: 64
                    height: 24
                }

                Label {
                    x: 3
                    y: 83
                    text: "Please note that the order in which EPI files are chosen can make a (small) difference in the resulting voxel displacement map"
                    font.italic: true
                }
                TextArea {
                    id: estimateText
                    x: 10
                    y: 260
                    width: 800
                    height: 231
                    //placeholderText: "Usage: estimate distortion
                    text: "### Usage: estimate distortion\n\n" +
"**Select FMRs/DMRs:** Choose the EPI files that are phase encoded either in opposite anterior-posterior (A>>P and P>>A) directions " +
"or opposite left-right directions (L>>R and R>>L) using the Browse button on the first two lines. " +
"Afterwards, in each dataset, select the volume which will be used to estimate the voxel displacement map (VDM). " +
"For diffusion-weighted data (spin echo EPI), a b=0 volume is usually best. " +
"The voxel displacement map (VDM) will be saved in *.map file format, which can be overlaid on FMR files; " +
"this may be useful to see the algorithm's result." +
"\n\n" +
"**Select volume numbers:** It is recommended to choose volumes acquired shortly after one another in time to minimize motion." +
"Ideally you would aquire one volume in the opposite phase encoding direction directly before starting the run(s) of your experiment. " +
"Then you can use that volume and the first volume of your first run for the estimation. " +
"For example, if there are four runs, the optimal scanning order would be two runs, then the volume with opposite " +
"phase encoding, then the anatomical scan, finally the last two runs. " +
"Our advice would then be to align all runs to the last volume of the second run (via intra-session alignment)." +
"\n\n" +
"**Pre-processing:** The estimation is based on image intensities and structure on which most of the pre-processing " +
"steps have a negligible effect. The only thing to make sure is to not apply any smoothing to the volumes" +
"used for estimation."
                    font.pixelSize: Qt.platform.os === "osx" ? 12 : 10
                    wrapMode: TextEdit.WordWrap
                    textFormat: TextEdit.MarkdownText
                    color: Qt.rgba(0.8, 0.8, 0.8, 1.0)
                }
                Row {
                    id: buttonsRowLayout2
                    x: 850//20
                    y: grbxEstimateVDM.height
                    spacing: 5

                    Item { Layout.fillWidth: true }
                    Button {
                        id: quitBtn
                        text: "Close"
                        onClicked: BrainVoyager.FinishPlugin() // 'ClosePlugin' for pure qml scripts, but if qml dialog serves as frontend for plugin, 'FinishPlugin()' needs to be called!
                    }
                    Button {
                        text: "Estimate VDM"
                        onClicked: root.onEstimate()
                    }
                }
            }
        } // end Item estimateVDMTab
        Item {
            id: applyVDMTab

            GroupBox {
                id: grbxApplyVDM
                title: "Select the voxel displacement map to be applied to the data:"
                Layout.fillWidth: true
                x: 10
                y: 40
                width: 1021
                height: 111//parent.height//400

                Row {

                    x: 20
                    y: 25

                    TextField {
                        id: vdmFilenameEdit
                        placeholderText: "Name of voxel displacement map (*_VDM.map)"
                        width: 800
                    }
                    Button {
                        id: vdmBrowseButton
                        text: "Browse"
                        onClicked: vdmFilenameEdit.text = root.onGetFileName("*_vdm.map")
                    }

                }
            }

            GroupBox {
                id: grbxApplyNames
                title: "Apply distortion correction to the following EPI data (*.fmr, *.dmr):"
                x: 10
                y: 160
                width: 1021
                height: 300

                TextArea {

                    id: textEditApplyNames
                    //    text: "EPI 1.fmr/dmr.... \n...\n...\n...\n...\n...\n...\n"
                    width: parent.width
                    height: 211
                }

                Row {
                    id: buttonsRowLayout
                    x: 600//20
                    y: 350//grbxApplyNames.height - 100 //textEdit.height + 15
                    spacing: 5

                    Button {
                        id: addButton
                        text: "Add"
                        onClicked: textEditApplyNames.append(root.onGetFileName("*.fmr *.dmr"))
                    }
                    Button {
                        id: checkButton
                        text: "Check names"
                        onClicked: root.onCheck()
                    }
                    Item { Layout.fillWidth: true }
                    Button {
                        id: applyCloseButton
                        text: "Close"
                        onClicked: BrainVoyager.FinishPlugin()
                    }
                    Button {
                        id: applyRunButton
                        text: "Apply VDM"
                        onClicked: root.onApplyVDM()
                    }
                } // end RowLayOut
            }
        } // end applyVDMTab
        /*   temporarily off
            Item {
            id: helpTab

            BrainVoyager.PrintToLog($("textBrowser").source);
            var sample_path = BrainVoyager.PathToSampleData;
            var html_index = sample_path.replace("SampleData/", "Extensions/Plugins/CopePluginHelp/index.html");
            var html_index_file =  "file:///".concat(html_index);
            $("textBrowser").source = new QUrl(html_index_file);
            BrainVoyager.PrintToLog($("textBrowser").source);

        } */
    }

    // ----- end definition of GUI controls -----

    enum Option {
        ESTIMATE = 0, // auto increment
        APPLY,
        ESTIMATE_APPLY_SAME,
        ESTIMATE_APPLY_ANY
    }

    readonly property variant experiment: {
        "nrOfProjects": 0,
        "option": 0,
        "arrayOfProjectParams": []
    }

    Component.onCompleted: {
        console.log("Plugin ready"); // + "  parent width x height: " + parent.width + " " + parent.height)
        BrainVoyager.PrintToLog("COPE plugin> plugin started");
    }

    function onGetFileName(fileExt) {

        var filename = BrainVoyager.BrowseFile("Please select the filename", fileExt);
        return filename;
    }


    // Called for the Browse buttons in the Select file for estimation manually tab (prev: Estimate VDM tab).
    function onSelectInput(index,spinBox) {

        BrainVoyager.SetPluginStringParameter("task", "select_fmr_or_dmr");
        BrainVoyager.ExecutePlugin();
        var filename = BrainVoyager.GetPluginStringParameter("filename"); // do not use BrowseFile here, because the plugin reads the nr of vols
        if (filename.length > 1) {

            if (index === 0) {
                volumeSpinBox0.to = BrainVoyager.GetPluginIntParameter("nr_of_volumes");
                volumeSpinBox0.value = BrainVoyager.GetPluginIntParameter("display_volume");
            } else {
                volumeSpinBox1.to = BrainVoyager.GetPluginIntParameter("nr_of_volumes");
                volumeSpinBox1.value = BrainVoyager.GetPluginIntParameter("display_volume");
            }
        }
        return filename;
    }


    // Called for the Run button on the Estimate VDM tab.

    function onEstimate() {
        // open FMR/DMR -- the C++ code needs an open FMR/DMR to create .map files
        var file0 = filenameEdit0.text;
        BrainVoyager.OpenDocument(file0);

        // set parameters
        BrainVoyager.SetPluginStringParameter("file0", file0);
        BrainVoyager.SetPluginStringParameter("file1", filenameEdit1.text);
        BrainVoyager.SetPluginIntParameter("volume0", volumeSpinBox0.value - 1);
        BrainVoyager.SetPluginIntParameter("volume1", volumeSpinBox1.value - 1);
        BrainVoyager.SetPluginIntParameter("rllr", estimateRllrRadioButton.checked);
        BrainVoyager.SetPluginIntParameter("use_ic", seRadioButton.checked);
        BrainVoyager.SetPluginIntParameter("use_ssd", ssdRadioButton.checked);
        //BrainVoyager.SetPluginIntParameter("apply_to_input", $("applyToInput").checked);// obsolete in v1.1
        BrainVoyager.SetPluginIntParameter("apply_to_input", rbtnEstimateAndApply.checked);
        //BrainVoyager.SetPluginIntParameter("correct_motion", $("correctMotion").checked);
        BrainVoyager.SetPluginIntParameter("save_derivative", saveDerivative.checked);
        //   BrainVoyager.SetPluginIntParameter("fast_algorithm", $("linearAlgorithm").checked); // obsolete in v1.2.2
        BrainVoyager.SetPluginIntParameter("order", spbOrderGauss.value);

        // call C++ function
        BrainVoyager.SetPluginStringParameter("task", "estimate");
        BrainVoyager.ExecutePlugin();
    }


    function onCheck() {
        BrainVoyager.SetPluginStringParameter("files", textEditApplyNames.text);//placeholderText);//$("textEdit").plainText);
        BrainVoyager.SetPluginStringParameter("task", "check");
        BrainVoyager.ExecutePlugin();
    }


    function onApplyVDM() {

        BrainVoyager.PrintToLog("Starting to apply one or several VDMs to: ");
        var files_string = textEditApplyNames.text;
        var files = files_string.split("\n");
        if (files.length == 0) {
            BrainVoyager.PrintToLog("No files found.");
            return;
        }
        BrainVoyager.OpenDocument(files[0]);
        BrainVoyager.SetPluginIntParameter("rllr", estimateRllrRadioButton.checked);
        BrainVoyager.SetPluginStringParameter("vdm", vdmFilenameEdit.text);
        BrainVoyager.SetPluginStringParameter("files", files_string);
        BrainVoyager.SetPluginStringParameter("task", "apply");
        BrainVoyager.ExecutePlugin();
    }


    function readBatchFile(exp) {

        //   general format of filenames file:

        //  general parameters
        //      no of projects:
        //      option: 0/1/2/3 (=estimate, apply or both to same, or both to any in session)

        //   for each project:
        //      estimate files:
        //      [estimate file 0:]
        //      [estimate file 1:]
        //      volumes:
        //      [volume 0:]
        //      [volume 1:]
        //      vdm file:
        //      (“none” if option == 3)
        //      no of apply files: [n]
        //      [run 0:]
        //      [run 1:]
        //      [run n:]

        var projcounter, projfilecounter;
        var success = true;
        var textfilename;
        if (batchFileLabel.text.length > 1) {
            textfilename = batchFileLabel.text;
        } else {
            textfilename = BrainVoyager.BrowseFile("Please select the text file with filenames", "*.txt");
        }
        BrainVoyager.PrintToLog("Start reading: " + textfilename);
        var name = String(textfilename); // circumvention of problem with implicit strings
        BrainVoyager.SetPluginStringParameter("task", "readbatchfile");
        BrainVoyager.SetPluginStringParameter("batchfilename", name);
        BrainVoyager.ExecutePlugin();

        var batchfileinfo = BrainVoyager.GetPluginStringParameter("filenames");
        BrainVoyager.PrintToLog(batchfileinfo);
        // now convert the batchfileinfo to experiment array
        var line;
        var experimentArray = [];
        var stoppoint = 0;
        var allLines = batchfileinfo.split('\n');//[0].split(':');
        for (var i=0; i<allLines.length; i++){
            if (allLines[i].indexOf('projects') > -1) {
                line = allLines[i].split(':');
                exp["noOfProjects"] = parseInt(line[1]);
            }
            if (allLines[i].indexOf('option') > -1) {
                line = allLines[i].split(':');
                exp["option"] = parseInt(line[1]);
                stoppoint = i;
            }
        } // stop looping

        for (projcounter = 0; projcounter < exp["noOfProjects"]; projcounter++) { // loop through projects
            var project = {};

            if ((exp["option"] === COPE_v122.Option.ESTIMATE) || (exp["option"] === COPE_v122.Option.ESTIMATE_APPLY_SAME)) {

                project.isnan = 0;
                project.estimateFile0 = allLines[(projcounter*7) + (stoppoint+3)].trim(); // estimate a voxel displacement map (VDM) from EPI and opposite phase encoded EPI file.
                project.estimateFile1 = allLines[(projcounter*7) + (stoppoint+4)].trim();
                project.estimateVolume0 = parseInt(allLines[(projcounter*7) + (stoppoint+6)]);
                project.estimateVolume1 = parseInt(allLines[(projcounter*7) + (stoppoint+7)]);

            } else if (exp["option"] === COPE_v122.Option.ESTIMATE_APPLY_ANY) { // this has variable length

                project.isnan = 0;
                project.estimateFile0 = allLines[stoppoint+3].trim(); // estimate a voxel displacement map (VDM) from EPI and opposite phase encoded EPI file.
                project.estimateFile1 = allLines[stoppoint+4].trim();
                project.estimateVolume0 = parseInt(allLines[stoppoint+6]);
                project.estimateVolume1 = parseInt(allLines[stoppoint+7]);
                project.vdmFile = allLines[stoppoint+9]; // should be none
                var applyFiles = [];
                line = allLines[(stoppoint+10)].split(':');
                project.nrOfApplyFiles = parseInt(line[1]);
                for (var applyfileNo = 0; applyfileNo < project.nrOfApplyFiles; applyfileNo++) {
                    applyFiles.push(allLines[(stoppoint+11) + applyfileNo].trim());
                }
                project.applyFiles = applyFiles;
                stoppoint += 10 + project.nrOfApplyFiles;

            } else if (exp["option"] === COPE_v122.Option.APPLY) { // OPTION 1

                project.isnan = 0;
                project.vdmFile = allLines[stoppoint+3].trim();
                var applyFiles = [];
                line = allLines[(stoppoint+4)].split(':');
                project.nrOfApplyFiles = parseInt(line[1]);

                for (var applyfileNo = 0; applyfileNo < project.nrOfApplyFiles; applyfileNo++) {
                    applyFiles.push(allLines[(stoppoint+5) + applyfileNo].trim());
                }
                project.applyFiles = applyFiles;
                stoppoint += 4 + project.nrOfApplyFiles;
            }

            experimentArray[projcounter] = project;
        } // end projectcounter

        exp.arrayOfProjectParams = experimentArray;
        return success;
    }

    function onStart() {

        var success = true;

        // print settings to BV Log tab
        BrainVoyager.PrintToLog("Phase encoding direction: " + (estimateRllrRadioButton.checked ? "RL-LR" : "AP-PA"));
        BrainVoyager.PrintToLog("Intensity correction: " + (seRadioButton.checked ? "yes" : "no"));
        BrainVoyager.PrintToLog("Similarity measure: " + (ssdRadioButton.checked ? "SSD" : "NCC"));
        BrainVoyager.PrintToLog("Save derivative: " + (saveDerivative.checked ? "yes" : "no"));
        BrainVoyager.PrintToLog("Order of Gaussian: " + spbOrderGauss.value);

        if (rbtnSingleProject.checked) {
            if ((rbtnEstimateAndApply.checked) || (rbtnApplyOnly.checked)) {
                root.onEstimate();
            } else if (rbtnApplyOnly.checked) {
                root.onApplyVDM();
            }
        } else if (rbtnBatch.checked) {
            success = root.readBatchFile(experiment);
            if (success) {
                if (experiment.option === COPE_v122.Option.ESTIMATE){
                    root.onEstimateBatch(experiment);
                } else if (experiment.option === COPE_v122.Option.ESTIMATE_APPLY_SAME) {
                    root.estimateApplyToSameBatch(experiment);
                } else if (experiment.option === COPE_v122.Option.APPLY) {
                    root.applyBatch(experiment);
                } else if (experiment.option === COPE_v122.Option.ESTIMATE_APPLY_ANY) {
                    root.onEstimateBatch(experiment); // this should create VDM files
                    root.applyBatch(experiment); // now apply VDM files to EPIs from same session
                }
            }
        }
      //  root.close(); // does not work
        BrainVoyager.FinishPlugin();

    }

    // Called for by the first tab "several files" option; including "estimate and apply" checkbox
    function estimateApplyToSameBatch(exp) {

        // open FMR/DMR -- the C++ code needs an open FMR/DMR to create .map files
        for (var i=0; i < exp.noOfProjects; i++) {

            BrainVoyager.PrintToLog("Processing project: " + (i+1));
            var projectParams = exp.arrayOfProjectParams[i];
            BrainVoyager.OpenDocument(projectParams.estimateFile0);
            BrainVoyager.SetPluginStringParameter("file0", projectParams.estimateFile0);
            BrainVoyager.SetPluginStringParameter("file1", projectParams.estimateFile1);
            BrainVoyager.SetPluginIntParameter("volume0", projectParams.estimateFileVolume0);
            BrainVoyager.SetPluginIntParameter("volume1", projectParams.estimateFileVolume1);
            BrainVoyager.SetPluginIntParameter("rllr", estimateRllrRadioButton.checked);
            BrainVoyager.SetPluginIntParameter("use_ic", seRadioButton.checked);
            BrainVoyager.SetPluginIntParameter("use_ssd", ssdRadioButton.checked);
            BrainVoyager.SetPluginIntParameter("apply_to_input", rbtnEstimateAndApply.checked);
            //BrainVoyager.SetPluginIntParameter("correct_motion", $("correctMotion").checked);
            BrainVoyager.SetPluginIntParameter("save_derivative", saveDerivative.checked);
            //    BrainVoyager.SetPluginIntParameter("fast_algorithm", $("linearAlgorithm").checked);
            BrainVoyager.SetPluginIntParameter("order", spbOrderGauss.value); //// CHECK!! -1? // 05-06-18

            // call C++ function
            BrainVoyager.SetPluginStringParameter("task", "estimate");
            BrainVoyager.ExecutePlugin();

            root.closeAllProjects();
        }
    }

    function applyBatch(exp) {

       // BrainVoyager.PrintToLog("Apply correction to batch:\n" + JSON.stringify(exp));
        for (var i=0; i < exp.noOfProjects; i++) {

            var projectParams = exp.arrayOfProjectParams[i];
            var files = projectParams.applyFiles;
            if (files.length === 0) {
                BrainVoyager.PrintToLog("No files to apply distortion correction to listed.");
                return;
            }

            if (projectParams.isnan !== 1) {
                BrainVoyager.OpenDocument(files[0].trim()); // 17-10-19 UIT TIJDELIJK
                BrainVoyager.SetPluginIntParameter("rllr", estimateRllrRadioButton.checked);
                BrainVoyager.SetPluginStringParameter("vdm", projectParams.vdmFile);
                var res = projectParams.applyFiles;
                var res2 = String(res).replace(/(,)/gm,"\n");
                BrainVoyager.SetPluginStringParameter("files", res2);
                BrainVoyager.SetPluginStringParameter("task", "apply");
                BrainVoyager.ExecutePlugin();
                this.closeAllProjects();
            } else {
                BrainVoyager.PrintToLog("No voxel displacement map for this dataset; please re-estimate.");
            }
        }
    }

    function onEstimateBatch(exp) {

        // open FMR/DMR -- the C++ code needs an open FMR/DMR to create .map files
        for (var i=0; i < exp["noOfProjects"]; i++) {

            BrainVoyager.PrintToLog("Estimating voxel displacement map for project: " + (i+1));
            var projectParams = exp["arrayOfProjectParams"][i];
            BrainVoyager.OpenDocument(projectParams.estimateFile0);
            BrainVoyager.SetPluginStringParameter("file0", projectParams.estimateFile0);
            BrainVoyager.PrintToLog("File 1: " + projectParams.estimateFile0);
            BrainVoyager.SetPluginStringParameter("file1", projectParams.estimateFile1);
            BrainVoyager.PrintToLog("File 2: " + projectParams.estimateFile1);
            BrainVoyager.SetPluginIntParameter("volume0", projectParams.estimateFileVolume0);
            BrainVoyager.SetPluginIntParameter("volume1", projectParams.estimateFileVolume1);
            BrainVoyager.SetPluginIntParameter("rllr", estimateRllrRadioButton.checked);
            BrainVoyager.SetPluginIntParameter("use_ic", seRadioButton.checked);
            BrainVoyager.SetPluginIntParameter("use_ssd", ssdRadioButton.checked);
            BrainVoyager.SetPluginIntParameter("apply_to_input", (exp.option === COPE_v122.Option.ESTIMATE_APPLY_SAME));// if apply to same: this will be performed straight away in plugin
            //BrainVoyager.SetPluginIntParameter("correct_motion", correctMotion.checked);
            BrainVoyager.SetPluginIntParameter("save_derivative", saveDerivative.checked);
            //    BrainVoyager.SetPluginIntParameter("fast_algorithm", linearAlgorithm.checked);
            BrainVoyager.SetPluginIntParameter("order", spbOrderGauss.value);

            // call C++ function
            BrainVoyager.SetPluginStringParameter("task", "estimate");
            BrainVoyager.ExecutePlugin();
            projectParams.vdmFile = BrainVoyager.GetPluginStringParameter("vdm");
            projectParams.isnan = BrainVoyager.GetPluginIntParameter("isnan");
            if (projectParams.isnan === 0) {
                BrainVoyager.PrintToLog("Name of VDM: " + projectParams.vdmFile); // fixed in plugin 12-07-18
            } else {
                projectParams.isnan = 1;
            }
            exp.arrayOfProjectParams[i] = projectParams; // now updated with VDM filename
            BrainVoyager.PrintToLog("End of estimating voxel displacement map for project: " + (i+1));
            root.closeAllProjects();
        }
    }

    function closeAllProjects() {

        var doc, i;
        var pn = BrainVoyager.PrintToLog;
        BrainVoyager.PrintToLog("Closing all projects.");
        var docs = BrainVoyager.Documents;
        var n_docs = docs.Count;

        if (n_docs === 0)
            return;

        for (i=n_docs; i>0; i--)
        {
            docs = BrainVoyager.Documents;
            n_docs = docs.Count;
            doc = BrainVoyager.ActiveDocument;
            doc.Close();
        }
    }
}
